@@ -47,29 +47,31 @@ static bool   byte_array_get_prop      (JSContext    *context,
 static bool   byte_array_set_prop      (JSContext    *context,
                                         JS::HandleObject obj,
                                         JS::HandleId id,
-                                        bool                   strict,
-                                        JS::MutableHandleValue value_p);
+                                        JS::MutableHandleValue value_p,
+                                        JS::ObjectOpResult&    result);
 GJS_NATIVE_CONSTRUCTOR_DECLARE(byte_array);
 static void   byte_array_finalize      (JSFreeOp     *fop,
                                         JSObject     *obj);
 
 static JSObject *gjs_byte_array_get_proto(JSContext *);
 
-struct JSClass gjs_byte_array_class = {
-    "ByteArray",
-    JSCLASS_HAS_PRIVATE |
-    JSCLASS_BACKGROUND_FINALIZE |
-    JSCLASS_IMPLEMENTS_BARRIERS,
+static const struct JSClassOps gjs_byte_array_class_ops = {
     NULL,  /* addProperty */
     NULL,  /* deleteProperty */
-    (JSPropertyOp)byte_array_get_prop,
-    (JSStrictPropertyOp)byte_array_set_prop,
+    byte_array_get_prop,
+    byte_array_set_prop,
     NULL,  /* enumerate */
     NULL,  /* resolve */
-    NULL,  /* convert */
+    nullptr,  /* mayResolve */
     byte_array_finalize
 };
 
+struct JSClass gjs_byte_array_class = {
+    "ByteArray",
+    JSCLASS_HAS_PRIVATE | JSCLASS_BACKGROUND_FINALIZE,
+    &gjs_byte_array_class_ops
+};
+
 bool
 gjs_typecheck_bytearray(JSContext       *context,
                         JS::HandleObject object,
@@ -199,7 +201,7 @@ byte_array_get_prop(JSContext *context,
 
     priv = priv_from_js(context, obj);
 
-    if (priv == NULL)
+    if (!priv)
         return true; /* prototype, not an instance. */
 
     JS::RootedValue id_value(context);
@@ -229,7 +231,7 @@ byte_array_length_getter(JSContext *context,
     GJS_GET_PRIV(context, argc, vp, args, to, ByteArrayInstance, priv);
     gsize len = 0;
 
-    if (priv == NULL)
+    if (!priv)
         return true; /* prototype, not an instance. */
 
     if (priv->array != NULL)
@@ -248,7 +250,7 @@ byte_array_length_setter(JSContext *context,
     GJS_GET_PRIV(context, argc, vp, args, to, ByteArrayInstance, priv);
     gsize len = 0;
 
-    if (priv == NULL)
+    if (!priv)
         return true; /* prototype, not instance */
 
     byte_array_ensure_array(priv);
@@ -268,7 +270,8 @@ byte_array_set_index(JSContext         *context,
                      JS::HandleObject obj,
                      ByteArrayInstance *priv,
                      gsize              idx,
-                     JS::MutableHandleValue value_p)
+                     JS::MutableHandleValue value_p,
+                     JS::ObjectOpResult&    result)
 {
     guint8 v;
 
@@ -289,7 +292,7 @@ byte_array_set_index(JSContext         *context,
     /* Stop JS from storing a copy of the value */
     value_p.setUndefined();
 
-    return true;
+    return result.succeed();
 }
 
 /* a hook on setting a property; set value_p to override property value to
@@ -299,15 +302,15 @@ static bool
 byte_array_set_prop(JSContext *context,
                     JS::HandleObject obj,
                     JS::HandleId id,
-                    bool strict,
-                    JS::MutableHandleValue value_p)
+                    JS::MutableHandleValue value_p,
+                    JS::ObjectOpResult&    result)
 {
     ByteArrayInstance *priv;
 
     priv = priv_from_js(context, obj);
 
-    if (priv == NULL)
-        return true; /* prototype, not an instance. */
+    if (!priv)
+        return result.succeed(); /* prototype, not an instance. */
 
     JS::RootedValue id_value(context);
     if (!JS_IdToValue(context, id, &id_value))
@@ -319,12 +322,12 @@ byte_array_set_prop(JSContext *context,
         if (!gjs_value_to_gsize(context, id_value, &idx))
             return false;
 
-        return byte_array_set_index(context, obj, priv, idx, value_p);
+        return byte_array_set_index(context, obj, priv, idx, value_p, result);
     }
 
     /* We don't special-case anything else for now */
 
-    return true;
+    return result.succeed();
 }
 
 static GByteArray *
@@ -385,7 +388,7 @@ byte_array_finalize(JSFreeOp *fop,
 
     priv = (ByteArrayInstance*) JS_GetPrivate(obj);
 
-    if (priv == NULL)
+    if (!priv)
         return; /* prototype, not instance */
 
     if (priv->array) {
@@ -405,30 +408,26 @@ to_string_func(JSContext *context,
                JS::Value *vp)
 {
     GJS_GET_PRIV(context, argc, vp, argv, to, ByteArrayInstance, priv);
-    char *encoding;
+    GjsAutoJSChar encoding;
     bool encoding_is_utf8;
     gchar *data;
 
-    if (priv == NULL)
+    if (!priv)
         return true; /* prototype, not instance */
 
     byte_array_ensure_array(priv);
 
     if (argc >= 1 && argv[0].isString()) {
-        if (!gjs_string_to_utf8(context, argv[0], &encoding))
+        JS::RootedString str(context, argv[0].toString());
+        encoding = JS_EncodeStringToUTF8(context, str);
+        if (!encoding)
             return false;
 
         /* maybe we should be smarter about utf8 synonyms here.
          * doesn't matter much though. encoding_is_utf8 is
          * just an optimization anyway.
          */
-        if (strcmp(encoding, "UTF-8") == 0) {
-            encoding_is_utf8 = true;
-            g_free(encoding);
-            encoding = NULL;
-        } else {
-            encoding_is_utf8 = false;
-        }
+        encoding_is_utf8 = (strcmp(encoding, "UTF-8") == 0);
     } else {
         encoding_is_utf8 = true;
     }
@@ -443,8 +442,7 @@ to_string_func(JSContext *context,
         /* optimization, avoids iconv overhead and runs
          * libmozjs hardwired utf8-to-utf16
          */
-        return gjs_string_from_utf8(context, data, priv->array->len,
-                                    argv.rval());
+        return gjs_string_from_utf8_n(context, data, priv->array->len, argv.rval());
     } else {
         bool ok = false;
         gsize bytes_written;
@@ -461,7 +459,6 @@ to_string_func(JSContext *context,
                            NULL, /* bytes read */
                            &bytes_written,
                            &error);
-        g_free(encoding);
         if (u16_str == NULL) {
             /* frees the GError */
             gjs_throw_g_error(context, error);
@@ -496,7 +493,7 @@ to_gbytes_func(JSContext *context,
     JSObject *ret_bytes_obj;
     GIBaseInfo *gbytes_info;
 
-    if (priv == NULL)
+    if (!priv)
         return true; /* prototype, not instance */
 
     byte_array_ensure_gbytes(priv);
@@ -534,11 +531,11 @@ from_string_func(JSContext *context,
 {
     JS::CallArgs argv = JS::CallArgsFromVp (argc, vp);
     ByteArrayInstance *priv;
-    char *encoding;
+    GjsAutoJSChar encoding;
     bool encoding_is_utf8;
     JS::RootedObject obj(context, byte_array_new(context));
 
-    if (obj == NULL)
+    if (!obj)
         return false;
 
     priv = priv_from_js(context, obj);
@@ -555,20 +552,16 @@ from_string_func(JSContext *context,
     }
 
     if (argc > 1 && argv[1].isString()) {
-        if (!gjs_string_to_utf8(context, argv[1], &encoding))
+        JS::RootedString str(context, argv[1].toString());
+        encoding = JS_EncodeStringToUTF8(context, str);
+        if (!encoding)
             return false;
 
         /* maybe we should be smarter about utf8 synonyms here.
          * doesn't matter much though. encoding_is_utf8 is
          * just an optimization anyway.
          */
-        if (strcmp(encoding, "UTF-8") == 0) {
-            encoding_is_utf8 = true;
-            g_free(encoding);
-            encoding = NULL;
-        } else {
-            encoding_is_utf8 = false;
-        }
+        encoding_is_utf8 = (strcmp(encoding, "UTF-8") == 0);
     } else {
         encoding_is_utf8 = true;
     }
@@ -577,15 +570,15 @@ from_string_func(JSContext *context,
         /* optimization? avoids iconv overhead and runs
          * libmozjs hardwired utf16-to-utf8.
          */
-        char *utf8 = NULL;
-        if (!gjs_string_to_utf8(context,
-                                argv[0],
-                                &utf8))
+        JS::RootedString str(context, argv[0].toString());
+        GjsAutoJSChar utf8 = JS_EncodeStringToUTF8(context, str);
+        if (!utf8)
             return false;
 
         g_byte_array_set_size(priv->array, 0);
-        g_byte_array_append(priv->array, (guint8*) utf8, strlen(utf8));
-        g_free(utf8);
+        g_byte_array_append(priv->array,
+                            reinterpret_cast<const guint8*>(utf8.get()),
+                            strlen(utf8));
     } else {
         JSString *str = argv[0].toString();  /* Rooted by argv */
         GError *error = NULL;
@@ -623,7 +616,6 @@ from_string_func(JSContext *context,
             }
         }
 
-        g_free(encoding);
         if (encoded == NULL) {
             /* frees the GError */
             gjs_throw_g_error(context, error);
@@ -650,9 +642,10 @@ from_array_func(JSContext *context,
     ByteArrayInstance *priv;
     guint32 len;
     guint32 i;
+    bool is_array;
     JS::RootedObject obj(context, byte_array_new(context));
 
-    if (obj == NULL)
+    if (!obj)
         return false;
 
     priv = priv_from_js(context, obj);
@@ -663,7 +656,9 @@ from_array_func(JSContext *context,
     priv->array = gjs_g_byte_array_new(0);
 
     JS::RootedObject array_obj(context, &argv[0].toObject());
-    if (!JS_IsArrayObject(context, array_obj)) {
+    if (!JS_IsArrayObject(context, array_obj, &is_array))
+        return false;
+    if (!is_array) {
         gjs_throw(context,
                   "byteArray.fromArray() called with non-array as first arg");
         return false;
@@ -722,7 +717,7 @@ from_gbytes_func(JSContext *context,
     gbytes = (GBytes*) gjs_c_struct_from_boxed(context, bytes_obj);
 
     JS::RootedObject obj(context, byte_array_new(context));
-    if (obj == NULL)
+    if (!obj)
         return false;
     priv = priv_from_js(context, obj);
     g_assert (priv != NULL);
