@@ -54,39 +54,38 @@
 
 #include "console.h"
 #include "cjs/context.h"
-#include "cjs/jsapi-private.h"
+#include "cjs/context-private.h"
 #include "cjs/jsapi-wrapper.h"
 
 static void
-gjs_console_error_reporter(JSContext *cx, const char *message, JSErrorReport *report)
+gjs_console_print_error(JSErrorReport *report)
 {
-    int i, j, k, n;
-    char *prefix, *tmp;
-    const char *ctmp;
+    /* Code modified from SpiderMonkey js/src/jscntxt.cpp, js::PrintError() */
 
-    if (!report) {
-        fprintf(stderr, "%s\n", message);
-        return;
-    }
+    g_assert(report);
 
-    prefix = NULL;
+    char *prefix = nullptr;
     if (report->filename)
         prefix = g_strdup_printf("%s:", report->filename);
     if (report->lineno) {
-        tmp = prefix;
-        prefix = g_strdup_printf("%s%u: ", tmp ? tmp : "", report->lineno);
+        char *tmp = prefix;
+        prefix = g_strdup_printf("%s%u:%u ", tmp ? tmp : "", report->lineno,
+                                 report->column);
         g_free(tmp);
     }
     if (JSREPORT_IS_WARNING(report->flags)) {
-        tmp = prefix;
+        char *tmp = prefix;
         prefix = g_strdup_printf("%s%swarning: ",
                                  tmp ? tmp : "",
                                  JSREPORT_IS_STRICT(report->flags) ? "strict " : "");
         g_free(tmp);
     }
 
+    const char *message = report->message().c_str();
+
     /* embedded newlines -- argh! */
-    while ((ctmp = strchr(message, '\n')) != NULL) {
+    const char *ctmp;
+    while ((ctmp = strchr(message, '\n')) != 0) {
         ctmp++;
         if (prefix)
             fputs(prefix, stderr);
@@ -99,34 +98,89 @@ gjs_console_error_reporter(JSContext *cx, const char *message, JSErrorReport *re
         fputs(prefix, stderr);
     fputs(message, stderr);
 
-    if (!report->linebuf) {
-        fputc('\n', stderr);
-        goto out;
-    }
+    if (const char16_t* linebuf = report->linebuf()) {
+        size_t n = report->linebufLength();
 
-    /* report->linebuf usually ends with a newline. */
-    n = strlen(report->linebuf);
-    fprintf(stderr, ":\n%s%s%s%s",
-            prefix,
-            report->linebuf,
-            (n > 0 && report->linebuf[n-1] == '\n') ? "" : "\n",
-            prefix);
-    n = ((char*)report->tokenptr) - ((char*) report->linebuf);
-    for (i = j = 0; i < n; i++) {
-        if (report->linebuf[i] == '\t') {
-            for (k = (j + 8) & ~7; j < k; j++) {
-                fputc('.', stderr);
+        fputs(":\n", stderr);
+        if (prefix)
+            fputs(prefix, stderr);
+
+        for (size_t i = 0; i < n; i++)
+            fputc(static_cast<char>(linebuf[i]), stderr);
+
+        // linebuf usually ends with a newline. If not, add one here.
+        if (n == 0 || linebuf[n - 1] != '\n')
+            fputc('\n', stderr);
+
+        if (prefix)
+            fputs(prefix, stderr);
+
+        n = report->tokenOffset();
+        for (size_t i = 0, j = 0; i < n; i++) {
+            if (linebuf[i] == '\t') {
+                for (size_t k = (j + 8) & ~7; j < k; j++)
+                    fputc('.', stderr);
+                continue;
             }
-            continue;
+            fputc('.', stderr);
+            j++;
         }
-        fputc('.', stderr);
-        j++;
+        fputc('^', stderr);
     }
-    fputs("^\n", stderr);
- out:
+    fputc('\n', stderr);
+    fflush(stderr);
     g_free(prefix);
 }
 
+static void
+gjs_console_warning_reporter(JSContext *cx, JSErrorReport *report)
+{
+    gjs_console_print_error(report);
+}
+
+/* Based on js::shell::AutoReportException from SpiderMonkey. */
+class AutoReportException {
+    JSContext *m_cx;
+
+public:
+    explicit AutoReportException(JSContext *cx) : m_cx(cx) {}
+
+    ~AutoReportException() {
+        if (!JS_IsExceptionPending(m_cx))
+            return;
+
+        /* Get exception object before printing and clearing exception. */
+        JS::RootedValue v_exn(m_cx);
+        (void) JS_GetPendingException(m_cx, &v_exn);
+
+        JS::RootedObject exn(m_cx, &v_exn.toObject());
+        JSErrorReport *report = JS_ErrorFromException(m_cx, exn);
+        if (report) {
+            g_assert(!JSREPORT_IS_WARNING(report->flags));
+            gjs_console_print_error(report);
+        } else {
+            JS::RootedString message(m_cx, JS::ToString(m_cx, v_exn));
+            if (!message) {
+                g_printerr("(could not convert thrown exception to string)\n");
+            } else {
+                GjsAutoJSChar message_utf8 = JS_EncodeStringToUTF8(m_cx, message);
+                g_printerr("%s\n", message_utf8.get());
+            }
+        }
+
+        JS::RootedObject stack(m_cx, ExceptionStackOrNull(exn));
+        if (stack) {
+            GjsAutoChar stack_str = gjs_format_stack_trace(m_cx, stack);
+            if (!stack_str)
+                g_printerr("(Unable to print stack trace)\n");
+            else
+                g_printerr("%s", stack_str.get());
+        }
+
+        JS_ClearPendingException(m_cx);
+    }
+};
+
 #ifdef HAVE_READLINE_READLINE_H
 static bool
 gjs_console_readline(JSContext *cx, char **bufp, FILE *file, const char *prompt)
@@ -154,6 +208,44 @@ gjs_console_readline(JSContext *cx, char **bufp, FILE *file, const char *prompt)
 }
 #endif
 
+/* Return value of false indicates an uncatchable exception, rather than any
+ * exception. (This is because the exception should be auto-printed around the
+ * invocation of this function.)
+ */
+static bool
+gjs_console_eval_and_print(JSContext  *cx,
+                           const char *bytes,
+                           size_t      length,
+                           int         lineno)
+{
+    JS::CompileOptions options(cx);
+    options.setUTF8(true)
+           .setFileAndLine("typein", lineno);
+
+    JS::RootedValue result(cx);
+    if (!JS::Evaluate(cx, options, bytes, length, &result)) {
+        if (!JS_IsExceptionPending(cx))
+            return false;
+    }
+
+    gjs_schedule_gc_if_needed(cx);
+
+    if (result.isUndefined())
+        return true;
+
+    JS::RootedString str(cx, JS::ToString(cx, result));
+    if (!str)
+        return true;
+
+    char *display_str;
+    display_str = gjs_value_debug_string(cx, result);
+    if (display_str) {
+        g_fprintf(stdout, "%s\n", display_str);
+        g_free(display_str);
+    }
+    return true;
+}
+
 static bool
 gjs_console_interact(JSContext *context,
                      unsigned   argc,
@@ -161,8 +253,6 @@ gjs_console_interact(JSContext *context,
 {
     JS::CallArgs argv = JS::CallArgsFromVp(argc, vp);
     bool eof = false;
-    JS::RootedValue result(context);
-    JS::RootedString str(context);
     JS::RootedObject global(context, gjs_get_import_global(context));
     GString *buffer = NULL;
     char *temp_buf = NULL;
@@ -170,7 +260,7 @@ gjs_console_interact(JSContext *context,
     int startline;
     FILE *file = stdin;
 
-    JS_SetErrorReporter(JS_GetRuntime(context), gjs_console_error_reporter);
+    JS::SetWarningReporter(context, gjs_console_warning_reporter);
 
         /* It's an interactive filehandle; drop into read-eval-print loop. */
     lineno = 1;
@@ -195,44 +285,25 @@ gjs_console_interact(JSContext *context,
         } while (!JS_BufferIsCompilableUnit(context, global,
                                             buffer->str, buffer->len));
 
-        JS::CompileOptions options(context);
-        options.setUTF8(true)
-               .setFileAndLine("typein", startline);
-        if (!JS::Evaluate(context, global, options, buffer->str, buffer->len,
-                          &result)) {
+        bool ok;
+        {
+            AutoReportException are(context);
+            ok = gjs_console_eval_and_print(context, buffer->str, buffer->len,
+                                            startline);
+        }
+        g_string_free(buffer, true);
+
+        auto gjs_context = static_cast<GjsContext *>(JS_GetContextPrivate(context));
+        ok = _gjs_context_run_jobs(gjs_context) && ok;
+
+        if (!ok) {
             /* If this was an uncatchable exception, throw another uncatchable
              * exception on up to the surrounding JS::Evaluate() in main(). This
              * happens when you run gjs-console and type imports.system.exit(0);
              * at the prompt. If we don't throw another uncatchable exception
              * here, then it's swallowed and main() won't exit. */
-            if (!JS_IsExceptionPending(context)) {
-                argv.rval().set(result);
-                return false;
-            }
+            return false;
         }
-
-        gjs_schedule_gc_if_needed(context);
-
-        if (JS_GetPendingException(context, &result)) {
-            str = JS::ToString(context, result);
-            JS_ClearPendingException(context);
-        } else if (result.isUndefined()) {
-            goto next;
-        } else {
-            str = JS::ToString(context, result);
-        }
-
-        if (str) {
-            char *display_str;
-            display_str = gjs_value_debug_string(context, result);
-            if (display_str != NULL) {
-                g_fprintf(stdout, "%s\n", display_str);
-                g_free(display_str);
-            }
-        }
-
- next:
-        g_string_free(buffer, true);
     } while (!eof);
 
     g_fprintf(stdout, "\n");
@@ -240,6 +311,7 @@ gjs_console_interact(JSContext *context,
     if (file != stdin)
         fclose(file);
 
+    argv.rval().setUndefined();
     return true;
 }
 
